import gzip
import io
import os
import shutil
import random
import time
import msgpack
from builtins import str as text


def write_json(o, filename):
    with gzip.open(filename, "w") as fp:
        msgpack.dump(o, fp, encoding='utf-8')


def read_json(filename):
    with gzip.open(filename, "r") as fp:
        return msgpack.load(fp, encoding='utf-8')


# TODO(derpferd): Use the move function to prevent RACE on files
class GameDB(object):
    TOKEN_LEN = 8

    def __init__(self, root_dir):
        self.root_dir = root_dir
        self.data_dir = os.path.join(self.root_dir, "data")
        self.game_dir = os.path.join(self.root_dir, "games")
        self.schools_dir = os.path.join(self.root_dir, "schools")
        self.competitions_dir = os.path.join(self.root_dir, "competitions")
        self.__load()

    def __load(self):
        if not os.path.exists(self.root_dir):
            os.mkdir(self.root_dir)
        if not os.path.exists(self.game_dir):
            os.mkdir(self.game_dir)
        if not os.path.exists(self.data_dir):
            os.mkdir(self.data_dir)
        if not os.path.exists(self.schools_dir):
            os.mkdir(self.schools_dir)
        if not os.path.exists(self.competitions_dir):
            os.mkdir(self.competitions_dir)

    def __get_user_tokens(self):
        return os.listdir(self.data_dir)

    def __get_game_tokens(self):
        return os.listdir(self.game_dir)

    def __get_school_tokens(self):
        return os.listdir(self.schools_dir)

    def __get_comp_tokens(self):
        return os.listdir(self.competitions_dir)

    def __get_school_user_tokens(self, school_tk):
        if self.is_school_token(school_tk):
            return os.listdir(self.__get_dir_for_token(school_tk, "tokens"))
        return []

    def __get_user_game_tokens(self, token):
        if self.is_user_token(token):
            return os.listdir(self.__get_dir_for_token(token, "games"))
        return []

    def __get_new_token(self, tokens=None, prefix=""):

        def new_token():
            return prefix + "".join([random.choice("0123456789ABCDEF") for _ in range(self.TOKEN_LEN)])
        if not tokens:
            tokens = self.__get_user_tokens()
        token = new_token()
        while token in tokens:
            token = new_token()
        return token

    def __get_dir_for_token(self, token, fns=[]):
        """Get the file path for a given token and optionally an additional path after the token dir.

        Args:
            token (str): The user's token.
            fns (str or list): The file or files to add after the token directory.
        """
        if isinstance(fns, str):
            fns = [fns]
        else:
            assert isinstance(fns, list)
        if self.is_user_token(token):
            return os.path.join(self.data_dir, token, *fns)
        elif self.is_game_token(token):
            return os.path.join(self.game_dir, token, *fns)
        elif self.is_school_token(token):
            return os.path.join(self.schools_dir, token, *fns)
        elif self.is_comp_token(token):
            return os.path.join(self.competitions_dir, token, *fns)
        return None

    def __get_cur_code_for_token(self, token):
        pass

    def __get_next_code_for_token(self, token):
        pass

    def is_comp_token(self, token):
        if len(token) > 0 and token[0] == "P":
            # It is a competition token
            return token in self.__get_comp_tokens()
        return False

    def is_school_token(self, token):
        if len(token) > 0 and token[0] == "S":
            # It is a school token
            return token in self.__get_school_tokens()
        return False

    def is_user_token(self, token):
        return token in self.__get_user_tokens()

    def is_game_token(self, gtoken):
        gtks = self.__get_game_tokens()
        return gtoken in gtks

    def get_new_token(self, school_tk, _token=None):  # Don't use `_token` unless you know what you are doing.
        # Is name needed?
        assert self.is_school_token(school_tk)
        token = _token
        if _token is None:
            token = self.__get_new_token()

        # Touch the file
        with open(self.__get_dir_for_token(school_tk, ["tokens", token]), "w") as fp:
            pass

        # Create token dir
        os.mkdir(os.path.join(self.data_dir, token))
        os.mkdir(os.path.join(self.data_dir, token, "games"))
        return token

    def add_new_school(self, name="", _token=None):  # Don't use `_token` unless you know what you are doing.
        token = _token
        if _token is None:
            token = self.__get_new_token(self.__get_school_tokens(), prefix="S")

        os.mkdir(os.path.join(self.schools_dir, token))
        os.mkdir(os.path.join(self.schools_dir, token, "tokens"))

        with io.open(os.path.join(self.schools_dir, token, "name"), "w", encoding="utf8") as fp:
            fp.write(text(name))

        return token

    def add_new_competition(self, name=""):
        token = self.__get_new_token(self.__get_comp_tokens(), prefix="P")

        os.mkdir(os.path.join(self.competitions_dir, token))
        os.mkdir(os.path.join(self.competitions_dir, token, "schools"))

        with io.open(os.path.join(self.competitions_dir, token, "name"), "w", encoding="utf8") as fp:
            fp.write(text(name))

        return token

    def add_new_game(self, frames=None, per_player_data=None, player_tokens=None):
        if per_player_data is not None:
            assert player_tokens is None
        if player_tokens is not None:
            assert per_player_data is None

        token = self.__get_new_token(self.__get_game_tokens(), prefix="G")

        os.mkdir(os.path.join(self.game_dir, token))
        os.mkdir(os.path.join(self.game_dir, token, "players"))

        if frames is not None:
            self.save_game_frames(token, frames)

        if player_tokens is not None:
            for player in player_tokens:
                self.set_game_player(token, player)
        elif per_player_data is not None:
            for player, data in per_player_data.items():
                self.set_game_player(token, player, data)

        with open(os.path.join(self.game_dir, token, "ctime"), "w") as fp:
            fp.write(text(time.time()))

        return token

    def add_school_to_comp(self, ctoken, stoken):
        assert self.is_comp_token(ctoken)
        assert self.is_school_token(stoken)

        school_dir = self.__get_dir_for_token(ctoken, ["schools", stoken])
        if not os.path.exists(school_dir):
            os.mkdir(school_dir)

    # TODO(derpferd): add function to remove a school

    def set_comp_school_code(self, ctoken, stoken, code):
        assert self.is_comp_token(ctoken)
        assert self.is_school_token(stoken)

        school_dir = self.__get_dir_for_token(ctoken, ["schools", stoken])
        if not os.path.exists(school_dir):
            os.mkdir(school_dir)

        with io.open(os.path.join(school_dir, "code.lp"), "w", encoding="utf8") as fp:
            fp.write(text(code))

    # def set_token_for_comp(self, ctoken, utoken, stoken):
    #     assert self.is_comp_token(ctoken)
    #     assert self.is_user_token(utoken)
    #     assert self.is_school_token(stoken)
    #
    #     school_dir = self.__get_dir_for_token(ctoken, ["schools", stoken])
    #     if not os.path.exists(school_dir):
    #         os.mkdir(school_dir)
    #     with io.open(os.path.join(school_dir, "code.lp"), "w", encoding="utf8") as fp:
    #         code = self.get_code(utoken)
    #         assert code is not None
    #         fp.write(code)
    #     # with open(os.path.join(school_dir, "name"), "w") as fp:
    #     #     name = self.get_name(stoken)
    #     #     assert name is not None
    #     #     fp.write(name)

    def get_comp_code(self, ctoken, stoken):
        school_dir = self.__get_dir_for_token(ctoken, ["schools", stoken])
        if os.path.exists(os.path.join(school_dir, "code.lp")):
            with io.open(os.path.join(school_dir, "code.lp"), "r", encoding="utf8") as fp:
                return fp.read()
        else:
            return None

    def get_comp_tokens(self):
        return self.__get_comp_tokens()

    def get_comps_for_token(self, utoken):
        comps = []
        stoken = self.get_school_for_token(utoken)
        for comp in self.__get_comp_tokens():
            if stoken in self.get_schools_in_comp(comp):
                comps += [comp]
        return comps

    def get_schools_in_comp(self, ctoken):
        return os.listdir(self.__get_dir_for_token(ctoken, "schools"))

    def set_comp_avg_score(self, ctoken, stoken, score):
        school_dir = self.__get_dir_for_token(ctoken, ["schools", stoken])
        assert school_dir is not None
        with io.open(os.path.join(school_dir, "avg_score"), "w", encoding="utf8") as fp:
            fp.write(text(score))

    def get_comp_avg_score(self, ctoken, stoken):
        school_dir = self.__get_dir_for_token(ctoken, ["schools", stoken])
        if os.path.exists(os.path.join(school_dir, "avg_score")):
            with io.open(os.path.join(school_dir, "avg_score"), "r", encoding="utf8") as fp:
                return float(fp.read())
        else:
            return None

    def save_code(self, token, code, options=None):
        """Save a user's code under their token.

        Args:
            token (str): The user's token.
            code (str): The user's code.
            options (json-able object): The user's options.
        """
        assert os.path.exists(self.__get_dir_for_token(token))
        with io.open(self.__get_dir_for_token(token, "code.lp"), "w", encoding="utf8") as fp:
            fp.write(text(code))
        if options:
            write_json(options, self.__get_dir_for_token(token, "options.mp.gz"))

    def save_name(self, token, name):
        """Save a user's name under their token.

        Args:
            token (str): The user's token.
            name (str): The user's name.
        """
        assert os.path.exists(self.__get_dir_for_token(token))
        with io.open(self.__get_dir_for_token(token, "name"), "w", encoding="utf8") as fp:
            fp.write(text(name))

    def save_avg_score(self, token, score):
        """Save a user's average score.

        Args:
            token (str): The user's token.
            score (int): The user's average score.
        """
        assert os.path.exists(self.__get_dir_for_token(token))
        with io.open(self.__get_dir_for_token(token, "avg_score"), "w", encoding="utf8") as fp:
            fp.write(text(score))

    def save_game_frames(self, gtoken, frames):
        assert os.path.exists(self.__get_dir_for_token(gtoken))
        write_json(frames, self.__get_dir_for_token(gtoken, "frames.mp.gz"))

    def set_game_player(self, gtoken, token, data=None):
        assert os.path.exists(self.__get_dir_for_token(gtoken, "players"))
        assert self.is_user_token(token)
        assert os.path.exists(self.__get_dir_for_token(token, "games"))

        os.mkdir(self.__get_dir_for_token(gtoken, ["players", token]))

        write_json(data, self.__get_dir_for_token(gtoken, ["players", token, "data.mp.gz"]))

        with open(self.__get_dir_for_token(token, ["games", gtoken]), "w"):
            pass

    def get_game_frames(self, gtoken):
        if os.path.exists(self.__get_dir_for_token(gtoken, "frames.mp.gz")):
            return read_json(self.__get_dir_for_token(gtoken, "frames.mp.gz"))

    def get_player_game_data(self, gtoken, token):
        if os.path.exists(self.__get_dir_for_token(gtoken, ["players", token, "data.mp.gz"])):
            return read_json(self.__get_dir_for_token(gtoken, ["players", token, "data.mp.gz"]))

    def get_games_for_token(self, token):
        return self.__get_user_game_tokens(token)

    def get_code_and_options(self, token):
        code, options = None, {}
        if os.path.exists(self.__get_dir_for_token(token, "code.lp")):
            with io.open(self.__get_dir_for_token(token, "code.lp"), "r", encoding="utf8") as fp:
                code = fp.read()
        if os.path.exists(self.__get_dir_for_token(token, "options.mp.gz")):
            options = read_json(self.__get_dir_for_token(token, "options.mp.gz"))
        return code, options

    def get_name(self, token):
        if self.is_user_token(token) or self.is_school_token(token) or self.is_comp_token(token):
            if os.path.exists(self.__get_dir_for_token(token, "name")):
                with io.open(self.__get_dir_for_token(token, "name"), "r", encoding="utf8") as fp:
                    return fp.read()
        return None

    def get_avg_score(self, token):
        if os.path.exists(self.__get_dir_for_token(token, "avg_score")):
            with io.open(self.__get_dir_for_token(token, "avg_score"), "r", encoding="utf8") as fp:
                try:
                    # Try to convert to float
                    return float(fp.read())
                except ValueError as e:
                    # If failed return none
                    return None
        else:
            return None

    def get_school_for_token(self, token):
        for school in self.__get_school_tokens():
            if token in self.__get_school_user_tokens(school):
                return school
        return None

    # Get tokens that belong to a school
    def get_tokens_for_school(self, school_tk):
        return self.__get_school_user_tokens(school_tk)

    def get_school_tokens(self):
        return self.__get_school_tokens()

    def get_players_for_game(self, gtoken):
        # TODO: add a test for this method
        return os.listdir(self.__get_dir_for_token(gtoken, "players"))

    def delete_game(self, gtoken):
        # TODO: add a test for this method
        assert self.is_game_token(gtoken)
        for player in self.get_players_for_game(gtoken):
            os.remove(self.__get_dir_for_token(player, ["games", gtoken]))
        shutil.rmtree(self.__get_dir_for_token(gtoken))



from .Game import GridGame
from .Game import NonGridGame
from .Game import GameLanguage
from .Game import GameRunner
from .Game import run
from .Comp import Ranking
from .Panels import Map
from .Panels import Panel
from .Panels import MapPanel
from .Panels import MessagePanel
from .Panels import StatusPanel
from .Panels import PanelBorder
from .Panels import PanelPadding

from .Player import Player

from .version import version

__version__ = version



from collections import defaultdict
from copy import copy
from functools import reduce

DEFAULT_CHAR = ' '


def is_char(c):
    return type(c) == str and len(c) == 1


def is_coord(p):
    return type(p) == tuple and len(p) == 2 and type(p[0]) == int and type(p[1]) == int


class Map(object):
    """
    char_to_ps: This is a dictionary which has a character for the keys and the value for the key is a set of all the
                positions where that char is located.
    p_to_char:  The is a dictionary which has a position(tuple with two elements: x and y) as a key and the current
                ColoredChar as the value.
    """

    def __init__(self, width, height, default_char=DEFAULT_CHAR):
        self.w = width
        self.h = height
        self.default_char = default_char
        self.char_to_ps = defaultdict(set)
        self.p_to_char = defaultdict(lambda: self.default_char)
        self.changes = {}

    def __setitem__(self, key, value):
        self.add(value, key)

    def __getitem__(self, item):
        return self.get_char_at(item)

    def get_x_y_dist_to_foo(self, pos, foo, wrapping=False, diagonal_moving=False, default=None):
        if type(pos) == list:
            pos = tuple(pos)
        assert is_coord(pos), "pos is not a coord: %r" % pos
        assert is_char(foo), "foo is not a char: %r" % foo

        dists = []
        x, y = pos
        for foo_pos in self.get_all_pos(foo):
            foo_x, foo_y = foo_pos
            # TODO(derpferd): Implement wrapping
            if wrapping or diagonal_moving:
                # for i, j in product(range(-1, 0, 1), range(-1, 0, 1)):
                #     pass
                raise NotImplementedError("Wrapping and Diagonal_moving not implemented")
            else:
                d_x, d_y = foo_x - x, foo_y - y
                dists += [(d_x, d_y)]
        if len(dists) == 0:
            return default
        else:
            return min(dists, key=lambda d: abs(d[0]) + abs(d[1]))

    # checks if pos is in bound of the map
    def in_bounds(self, pos):
        assert is_coord(pos), "pos: '{}', isn't a coord".format(pos)
        x, y = pos
        return 0 <= x < self.w and 0 <= y < self.h

    def wrap(self, pos, wrap_x=False, wrap_y=False):
        x, y = pos
        if wrap_x:
            x %= self.w
        if wrap_y:
            y %= self.h
        return (x, y)

    # changes in the format of a dictionary
    # key: (x, y)
    # value: new_char
    def get_diff(self):
        changes = self.changes
        self.changes = {}
        return changes

    # pos must be tuple
    def add(self, char, pos):
        assert is_char(char), "'{}' isn't a char".format(char)
        assert self.in_bounds(pos), "'{}' isn't in bounds".format(pos)

        if pos in self.p_to_char.keys():
            self.rm_char(pos)
        self.char_to_ps[char].add(pos)
        self.p_to_char[pos] = char
        self.changes[pos] = char

    # pos must be tuple
    def rm_char(self, pos):
        assert self.in_bounds(pos)
        char = self.p_to_char[pos]
        del self.p_to_char[pos]
        if char in self.char_to_ps and pos in self.char_to_ps[char]:
            self.char_to_ps[char].remove(pos)
            self.changes[pos] = self.default_char

    # returns a set of pos
    def get_all_pos(self, char):
        assert is_char(char)
        return copy(self.char_to_ps[char])

    # will return default_char if the position is not set
    def get_char_at(self, pos):
        assert self.in_bounds(pos)
        return copy(self.p_to_char[pos])

    # offset should be in the format tuple(x, y)
    def shift_all(self, offset, wrap_x=False, wrap_y=False):
        assert type(offset) == tuple
        old_char_to_ps = copy(self.char_to_ps)
        self.char_to_ps = defaultdict(set)
        self.p_to_char = defaultdict(lambda: self.default_char)
        for char, ps in old_char_to_ps.items():
            for p in ps:
                if p not in self.changes:
                    self.changes[p] = self.default_char
                new_pos = self.wrap((p[0] + offset[0], p[1] + offset[1]), wrap_x, wrap_y)
                if self.in_bounds(new_pos):
                    self.add(char, new_pos)


class PanelPadding(object):
    TOP = 1
    RIGHT = 2
    LEFT = 4
    BOTTOM = 8
    DEFAULT_SIZE = 0

    def __init__(self, sides=0, size=0):
        self.sides = sides
        self.style = {self.sides: size}

    @staticmethod
    def create(top=False, bottom=False, right=False, left=False):
        paddings = [PanelPadding()]
        if type(top) == bool and top:
            paddings += [PanelPadding(PanelPadding.TOP)]
        elif type(top) == int:
            paddings += [PanelPadding(PanelPadding.TOP, top)]
        if type(bottom) == bool and bottom:
            paddings += [PanelPadding(PanelPadding.BOTTOM)]
        elif type(bottom) == int:
            paddings += [PanelPadding(PanelPadding.BOTTOM, bottom)]
        if type(right) == bool and right:
            paddings += [PanelPadding(PanelPadding.RIGHT)]
        elif type(right) == int:
            paddings += [PanelPadding(PanelPadding.RIGHT, right)]
        if type(left) == bool and left:
            paddings += [PanelPadding(PanelPadding.LEFT)]
        elif type(left) == int:
            paddings += [PanelPadding(PanelPadding.LEFT, left)]

        return reduce(lambda x, y: x | y, paddings)

    def __or__(self, other):
        assert isinstance(other, PanelPadding)
        new_padding = PanelPadding()
        new_padding.sides = other.sides | self.sides
        new_padding.style = self.style
        new_padding.style.update(other.style)
        return new_padding

    def __getitem__(self, item):
        if item in self.style:
            if self.style[item]:
                return self.style[item]
        return self.DEFAULT_SIZE

    def __contains__(self, item):
        return self.has_side(item)

    def has_side(self, side):
        return True
        # return side in self.style


class PanelBorder(object):
    TOP = 1
    RIGHT = 2
    LEFT = 4
    BOTTOM = 8
    DEFAULT_CHARS = {TOP: chr(196), BOTTOM: chr(196), RIGHT: chr(179), LEFT: chr(179),
                     TOP | RIGHT: chr(191), TOP | LEFT: chr(218), BOTTOM | RIGHT: chr(217), BOTTOM | LEFT: chr(192)}

    def __init__(self, sides=0, char=None):
        self.sides = sides
        self.style = {self.sides: char}

    @staticmethod
    def create(top=False, bottom=False, right=False, left=False, top_right=False, top_left=False, bottom_right=False,
               bottom_left=False):
        """
        Use this method for an easy way to create nice borders. Set any side to true to create a border with the default
        characters. You can also set any side to a character which will create a border with that character instead.
        In order for the corners (top_right, top_left, ...) to be rendered both of their respective sides must be set
        to True or a char.

        Ex.
        >>> PanelBorder.create(top=True)  # This will create a top border with the default char

        >>> PanelBorder.create(top='$')  # This will create a top border of '$'s.

        >>> PanelBorder.create(top='=', left='!', top_left='?')  # This will create a border looking like some thing below.
            ?====
            !
            !
        """
        borders = []
        if type(top) == bool and top:
            borders += [PanelBorder(PanelBorder.TOP)]
        elif type(top) == str:
            borders += [PanelBorder(PanelBorder.TOP, char=top)]
        if type(bottom) == bool and bottom:
            borders += [PanelBorder(PanelBorder.BOTTOM)]
        elif type(bottom) == str:
            borders += [PanelBorder(PanelBorder.BOTTOM, char=bottom)]
        if type(right) == bool and right:
            borders += [PanelBorder(PanelBorder.RIGHT)]
        elif type(right) == str:
            borders += [PanelBorder(PanelBorder.RIGHT, char=right)]
        if type(left) == bool and left:
            borders += [PanelBorder(PanelBorder.LEFT)]
        elif type(left) == str:
            borders += [PanelBorder(PanelBorder.LEFT, char=left)]
        if type(top_right) == bool and top_right:
            borders += [PanelBorder(PanelBorder.TOP | PanelBorder.RIGHT)]
        elif type(top_right) == str:
            borders += [PanelBorder(PanelBorder.TOP | PanelBorder.RIGHT, char=top_right)]
        if type(top_left) == bool and top_left:
            borders += [PanelBorder(PanelBorder.TOP | PanelBorder.LEFT)]
        elif type(top_left) == str:
            borders += [PanelBorder(PanelBorder.TOP | PanelBorder.LEFT, char=top_left)]
        if type(bottom_right) == bool and bottom_right:
            borders += [PanelBorder(PanelBorder.BOTTOM | PanelBorder.RIGHT)]
        elif type(bottom_right) == str:
            borders += [PanelBorder(PanelBorder.BOTTOM | PanelBorder.RIGHT, char=bottom_right)]
        if type(bottom_left) == bool and bottom_left:
            borders += [PanelBorder(PanelBorder.BOTTOM | PanelBorder.LEFT)]
        elif type(bottom_left) == str:
            borders += [PanelBorder(PanelBorder.BOTTOM | PanelBorder.LEFT, char=bottom_left)]

        return reduce(lambda x, y: x | y, borders)

    def __or__(self, other):
        assert isinstance(other, PanelBorder)
        new_border = PanelBorder()
        new_border.sides = other.sides | self.sides
        new_border.style = self.style
        new_border.style.update(other.style)
        return new_border

    def __getitem__(self, item):
        if item in self.style:
            if self.style[item]:
                return self.style[item]
        return self.DEFAULT_CHARS[item]

    def __contains__(self, item):
        return self.has_side(item)

    def has_side(self, side):
        return side in self.style


class Panel(Map):
    def __init__(self, x, y, w, h, background_char=DEFAULT_CHAR, border=PanelBorder(), padding=PanelPadding()):
        super(Panel, self).__init__(w, h, background_char)
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.border = border
        self.padding = padding
        self.real_y = y
        self.real_x = x
        self.real_w = w
        self.real_h = h
        if PanelBorder.TOP in self.border:
            self.real_y -= 1
            self.real_h += 1
        if PanelBorder.LEFT in self.border:
            self.real_x -= 1
            self.real_w += 1
        if PanelBorder.BOTTOM in self.border:
            self.real_h += 1
        if PanelBorder.RIGHT in self.border:
            self.real_w += 1
        if PanelPadding.TOP in self.padding:
            self.y += self.padding[PanelPadding.TOP]
            self.h -= self.padding[PanelPadding.TOP]
        if PanelPadding.LEFT in self.padding:
            self.x += self.padding[PanelPadding.LEFT]
            self.w -= self.padding[PanelPadding.LEFT]
        if PanelPadding.BOTTOM in self.padding:
            self.h -= self.padding[PanelPadding.BOTTOM]
        if PanelPadding.RIGHT in self.padding:
            self.w -= self.padding[PanelPadding.RIGHT]

    def draw_char(self, char, pos, frame_buffer):
        # TODO: add asserts for these
        x, y = pos
        assert is_char(char)
        frame_buffer.set(x, y, char)

    def redraw(self, frame_buffer):
        try:
            if PanelBorder.TOP in self.border:
                for i in range(self.real_w):
                    frame_buffer.set(self.real_x + i, self.real_y, self.border[PanelBorder.TOP])
            if PanelBorder.LEFT in self.border:
                for i in range(self.real_h):
                    frame_buffer.set(self.real_x, self.real_y + i, self.border[PanelBorder.LEFT])
            if PanelBorder.BOTTOM in self.border:
                for i in range(self.real_w):
                    frame_buffer.set(self.real_x + i, self.real_y + self.real_h - 1,
                                     self.border[PanelBorder.BOTTOM])
            if PanelBorder.RIGHT in self.border:
                for i in range(self.real_h):
                    frame_buffer.set(self.real_x + self.real_w - 1, self.real_y + i,
                                     self.border[PanelBorder.RIGHT])
            if PanelBorder.TOP in self.border and PanelBorder.LEFT in self.border:
                frame_buffer.set(self.real_x, self.real_y, self.border[PanelBorder.TOP | PanelBorder.LEFT])
            if PanelBorder.BOTTOM in self.border and PanelBorder.LEFT in self.border:
                frame_buffer.set(self.real_x, self.real_y + self.real_h - 1,
                                 self.border[PanelBorder.BOTTOM | PanelBorder.LEFT])
            if PanelBorder.TOP in self.border and PanelBorder.RIGHT in self.border:
                frame_buffer.set(self.real_x + self.real_w - 1, self.real_y,
                                 self.border[PanelBorder.TOP | PanelBorder.RIGHT])
            if PanelBorder.BOTTOM in self.border and PanelBorder.RIGHT in self.border:
                frame_buffer.set(self.real_x + self.real_w - 1, self.real_y + self.real_h - 1,
                                 self.border[PanelBorder.BOTTOM | PanelBorder.RIGHT])
        except IndexError:
            raise IndexError("Out of bounds while drawing border. This is normally caused by the forgetting that the"
                             "x,y adn width, height of a Panel doesn't include the border.")


class MapPanel(Panel):
    def __init__(self, x, y, w, h, default_char=DEFAULT_CHAR, border=PanelBorder(), padding=PanelPadding()):
        """

        Args:
            x: The x value of the position of top left corner (not including the border)
            y: The y value of the position of top left corner (not including the border)
            w: The width of the panel not including the border
            h: The height of the panel not including the border
            default_char:
            border:
            padding:
        """
        super(MapPanel, self).__init__(x, y, w, h, default_char, border, padding)
        self.is_first = True

    def first_draw(self, frame_buffer):
        # pass
        for x in range(self.w):
            for y in range(self.h):
                self.draw_char(self[(x, y)], (self.x + x, self.y + y), frame_buffer)

    def redraw(self, frame_buffer):
        super(MapPanel, self).redraw(frame_buffer)
        if self.is_first:
            self.first_draw(frame_buffer)
            self.is_first = False
        diff = self.get_diff()
        for pos in diff:
            char = diff[pos]

            # Check that the position is in bounds
            if 0 <= pos[0] < self.w and 0 <= pos[1] < self.h:
                self.draw_char(char, (self.x + pos[0], self.y + pos[1]), frame_buffer)
            else:
                raise Warning("Char out of bounds: Decided to skip drawing it!")


class MessagePanel(Panel):
    """ This panel contains messages to be displayed to the user. It acts as a scrolling text log.
        This means that the first message is display on the first line and some on till you have
        enough messages to fill the height. Then the next message added will push all the other
        messages up and the first message will be discarded.

        Example:
            A MessagePanel with height of 3. Add messages "A", "B" and "C". It will display like the following:
                A
                B
                C
            Add message "D" and it will display like the following:
                B
                C
                D

        Note:
            The following is deprecated:
            >>> m = MessagePanel()
            >>> m += ["Hi"]
            You should do this instead:
            >>> m = MessagePanel()
            >>> m.add("Hi")

    """
    def __init__(self, x, y, w, h, default_char=DEFAULT_CHAR, border=PanelBorder(),
                 padding=PanelPadding.create(*[1] * 4)):
        super(MessagePanel, self).__init__(x, y, w, h, default_char, border, padding)
        self.msgs = []
        self.rows = self.h
        self.max_len = self.w

    def __add__(self, other):
        print("The __add__ method on a MessagePanel has been deprecated.")
        self.add(other)
        return self

    def __getitem__(self, item):
        assert isinstance(item, int)
        return self.msgs[item]

    # TODO: Decide if this should be allowed
    # Message should act like a stack without popping only adding
    def __setitem__(self, key, value):
        assert isinstance(key, int)
        assert isinstance(value, str)
        assert "\n" not in value
        if 0 <= key < len(self.msgs):
            self.msgs[key] = value

    def __iter__(self):
        return self.msgs.__iter__()

    def __contains__(self, item):
        return item in self.msgs

    def add(self, messages):
        assert isinstance(messages, str) or hasattr(messages, '__iter__')
        if isinstance(messages, str):
            messages = [messages]
        for message in messages:
            self.msgs += message.split("\n")

    def clear(self):
        self.msgs = []

    def get_current_messages(self):
        return self.msgs[-self.rows:]

    def redraw(self, frame_buffer):
        super(MessagePanel, self).redraw(frame_buffer)
        msgs_to_display = self.get_current_messages()

        # Clear the WHOLE message panel
        for j in range(self.rows):
            # TODO: optimize this by only clearing what should be changed.
            # Clear msg row
            for i in range(self.max_len):
                frame_buffer.set(self.x + i, self.y + j, self.default_char)

        # Write the message we have to the panel
        for j in range(len(msgs_to_display)):
            msg = msgs_to_display[j]
            # Print msg
            for i in range(min(len(msg), self.max_len)):
                frame_buffer.set(self.x + i, self.y + j, msg[i])


class StatusPanel(MessagePanel):
    def __init__(self, x, y, w, h, default_char=DEFAULT_CHAR, border=PanelBorder(),
                 padding=PanelPadding().create(*[1] * 4)):
        super(StatusPanel, self).__init__(x, y, w, h, default_char, border, padding)
        self.info = {}

    def __getitem__(self, item):
        return self.info[item]

    def __setitem__(self, key, value):
        self.info[str(key)] = str(value)

    def __contains__(self, item):
        return item in self.info

    def get_current_messages(self):
        self.msgs = []
        for key in self.info:
            self.msgs += [key + ": " + self.info[key]]
        return super(StatusPanel, self).get_current_messages()



from __future__ import print_function

from random import random, choice, shuffle

from .Game import GameRunner
from .Player import Room


def create_room(gamedb, bot, compiler, size):
    stokens = gamedb.get_school_tokens()
    pool = []
    for stoken in stokens:
        pool += gamedb.get_tokens_for_school(stoken)

    if len(pool) == 0:
        from copy import deepcopy
        return Room([bot] + [deepcopy(bot) for _ in range(size - 1)])

    bots = [bot]
    while True:
        token = choice(pool)
        code, options = gamedb.get_code_and_options(token)
        try:
            prog = compiler.compile(code)
            prog.options = options
            bots += [prog]
            if len(bots) >= size:
                return Room(bots)
        except:
            print("Couldn't compile code for '{}' in '{}'".format(token, gamedb.get_school_for_token(token)))


def avg(scores):
    return float((sum(scores) * 100) / len(scores)) / 100


def sim_competition(compiler, game, gamedb, token, runs, debug=False, score_func=avg):
    assert gamedb is not None
    assert gamedb.is_comp_token(token)

    seeds = [random() for _ in xrange(2 * runs + 5)]

    for school in gamedb.get_schools_in_comp(token):
        if debug:
            print("Got school '" + school + "'")
        max_score = 0
        max_code = ""
        for student in gamedb.get_tokens_for_school(school):
            if debug:
                print("Got student '" + student + "'")
            code, options = gamedb.get_code_and_options(student)
            if not code:
                continue
            if debug:
                print("Compiling code...")
            prog = compiler.compile(code)
            prog.options = options
            if debug:
                print("setting up game runner...")
            runner = GameRunner(game, Room([prog]))
            if debug:
                print("Simulating...")
            score = None

            scores = []
            count = 0
            seed = 0
            # TODO: make this able to run in a pool of threads (so it can be run on multiple CPUs)
            while count < runs:
                try:
                    if seed >= len(seeds):
                        print("Ran out of seeds")
                        break
                    scores += [runner.run_for_avg_score(times=1, seed=seeds[seed])]
                    # print(scores[-1])
                    # import sys
                    # sys.stdout.flush()
                    count += 1
                    seed += 1
                except Exception as e:
                    print("There was an error simulating the game (Moving to next seed):", e)
                    seed += 1
            score = score_func(scores)
            # score = runner.run_for_avg_score(times=runs)

            # while score is None:
            #     try:
            #     except Exception as e:
            #         print("There was an error simulating the game:", e)
            if score > max_score:
                max_score = score
                max_code = code
        if debug:
            print("Saving score...", max_score)
        gamedb.set_comp_avg_score(token, school, max_score)
        gamedb.set_comp_school_code(token, school, max_code)
    if debug:
        print("All done :)")


class Ranking(object):
    def __init__(self, bots):
        """

        Args:
            bots: Tuple containing standing and bot id
        """
        self.ranks = {}
        for i in range(0, len(bots)):
            self.ranks[bots[i]] = i  # Key: BOTS Value: RANK

    def __add__(self, other):
        new_rank = {}
        for k in self.ranks:
            new_rank[k] = self.ranks[k] + other[k]
        return new_rank

    def __radd__(self, other):
        if other == 0:
            return self
        new_rank = {}
        for k in self.ranks:
            new_rank[k] = self.ranks[k] + other[k]
        return new_rank

    def add_rank(self, standing, bot):
        self.ranks[bot] += standing


class MultiplayerComp(object):
    RUN_FACTOR = 4

    def __init__(self, bots, room_size, default_bot_class):
        """

        Args:
            bots: (PROG) the players program to be executed and ranked
            room_size: (INT) the size of the rooms
        """
        self.default_bot_class = default_bot_class
        self.room_size = room_size
        self.scores = {}  # Prog:scores
        self.rooms = {}  # Rooms:Rankings
        self.cur_run = 0
        for bot in bots:
            self.scores[bot] = 0

    def __iter__(self):
        return self

    def __setitem__(self, key, value):
        """
        ARGS:
            key (Room): The room to set to ranking
            value (Ranking): The returned ranking
        """
        self.rooms[key] = value
        for k in value.ranks:
            if k.prog in self.scores:
                self.scores[k.prog] += value.ranks[k]

    def __next__(self):
        return self.next()

    def next(self):
        if self.cur_run == self.RUN_FACTOR:
            raise StopIteration()

        l = list(self.scores.keys())
        shuffle(l)
        p = l[:self.room_size + 1]
        while len(p) < self.room_size:
            p += [self.default_bot_class()]
        room = Room(p)
        self.rooms[room] = None
        self.cur_run += 1
        return room

    @staticmethod
    def sim_multiplayer(s_token, gamedb, game, compiler, debug=False):
        assert game.MULTIPLAYER, "Game must be multi-player to do MultiplayerComp."
        assert gamedb is not None
        assert gamedb.is_school_token(s_token)
        students = gamedb.get_tokens_for_school(s_token)  # Only getting one school token
        bots = []
        for s in students:
            try:
                if debug:
                    print("got student '" + s + "'")
                code, options = gamedb.get_code_and_options(s)
                name = gamedb.get_name(s)
                if not code:
                    continue
                if debug:
                    print("compiling code...")
                prog = compiler.compile(code)
                prog.options = options
                prog.token = s
                prog.name = name
                if debug:
                    print("simulating...")
                bots += [prog]
            except:
                print("Couldn't compile code for '{}' in '{}'".format(s, gamedb.get_school_for_token(s)))
        tourney = MultiplayerComp(bots, game.get_number_of_players(), game.default_prog_for_computer())
        for room in tourney:
            if debug:
                print("Room: " + str(room))
            gamerunner = GameRunner(game, room)
            tourney[room] = gamerunner.run_for_avg_score(times=1, func=sum)

        for player, score in tourney.scores.items():
            gamedb.save_avg_score(player.token, score)
            if debug:
                print("Score {} for Bike: {}".format(score, str(player.token)))



import sys
import traceback


# A template class for the Prog for the Player.
class Prog(object):
    def __init__(self):
        # The `options`, `token` and `name` attributes are reserved for the GameRunner
        self.options = {}
        self.token = None
        self.name = None

    def run(self, state=None, max_op_count=-1):
        pass


class UserProg(Prog):
    def __init__(self):
        super().__init__()
        self.name = "You"
        self.key = None

    def run(self, *args, **kwargs):
        return {"move": ord(self.key)}


class Player(object):
    def __init__(self, prog):
        self.prog = prog
        self.prev_vars = {}
        self.debugging = prog.options.get("debug", False)
        self.debug_vars = []  # TODO: Maybe make this into a class.

    def run_turn(self, random, max_ops=1000000):
        try:
            nxt_state = self.prog.run(self.get_state(), max_op_count=max_ops, random=random)
            self.update_state(dict(nxt_state))
            self.prev_vars = nxt_state
        except:
            traceback.print_exc(file=sys.stdout)
            print("Player class had an error!")
            self.update_state({})  # if the program had an error pass an empty state.
            self.prev_vars = {}

    def get_state(self):
        raise Exception("Not implemented!")


class DefaultGridPlayer(Player):
    """Make sure that all bot_vars are updated in game.do_turn"""
    def __init__(self, prog, bot_consts):
        """
        Args:
             prog(Prog):
             bot_consts(ConstMapping):
        """
        super(DefaultGridPlayer, self).__init__(prog)
        self.bot_consts = bot_consts
        self.bot_vars = {}
        self.move = None

    def get_state(self):
        state = dict(self.prev_vars)
        state.update(self.bot_consts)
        state.update(self.bot_vars)
        return state

    def update_state(self, state):
        # remove consts
        for key in self.bot_consts:
            state.pop(key, None)

        self.move = chr(state.get("move", ord("Q")))

        if self.debugging:
            human_vars = {}
            for name, val in state.items():
                if isinstance(val, int) and val in self.bot_consts:
                    human_vars[name] = self.bot_consts[val] + " (" + str(val) + ")"
                elif str(val) == str(True):
                    human_vars[name] = 1
                elif str(val) == str(False):
                    human_vars[name] = 0
                else:
                    human_vars[name] = val
            self.debug_vars += [human_vars]


class Room(object):
    def __init__(self, bots=None):
        if bots is None:
            self.bots = []
        else:
            self.bots = bots

        self.debug_vars = {}
        self.screen_cap = None

    def set_bot_debug(self, bot, debug_vars):
        self.debug_vars[bot] = debug_vars

    def set_playback(self, screen_cap):
        self.screen_cap = screen_cap



from copy import copy, deepcopy


class FrameBuffer(object):
    # TODO: decide what should go here.
    pass
    # def __init__(self, width, height):
    #     self.x = width
    #     self.y = height


class GridFrameBuffer(FrameBuffer):
    def __init__(self, width, height, charset=None, init_value='\x00'):
        """
        Args:
            width(int):
            height(int):
            charset(CharSet): The char set to use when drawing this frame buffer. (This is optional. If you don't set
                              then any draw function will not work.)
            init_value:
        """
        # super(GridFrameBuffer, self).__init__(width, height)
        self.x = width
        self.y = height
        self.charset = charset
        self.arr = []
        for i in range(height):
            self.arr += [[ord(init_value)]*width]

    @classmethod
    def from_string_array(cls, frame):
        assert len(frame), "Frame must contain something"
        x, y = len(frame[0]), len(frame)
        buf = cls(x, y)
        for i in range(x):
            for j in range(y):
                buf.set(i, j, ord(frame[j][i]))
        return buf

    def __eq__(self, other):
        if self.x != other.x or self.y != other.y:
            return False
        return self.arr == other.arr

    def __str__(self):
        return "\n".join(["".join(map(chr, x)) for x in self.arr])

    def __repr__(self):
        return "<GridFrameBuffer: '{}'".format(str(self).replace("\n", "\\n"))

    def set(self, x, y, char):
        """

        Args:
            x(int):
            y(int):
            char(int): The id of the char.
        """
        if isinstance(char, str):
            assert len(char) == 1
            char = ord(char)
        if not (0 <= x < self.x):
            raise IndexError("Trying to draw out of bounds at ({}, {})".format(x, y))
        if not (0 <= y < self.y):
            raise IndexError("Trying to draw out of bounds at ({}, {})".format(x, y))
        self.arr[y][x] = char

    def draw_to_surface(self, surface):
        if not self.charset:
            raise ValueError("CharSet must be set if you would like to draw anything!")
        # TODO: handle background better
        surface.fill((0, 0, 0))
        for y, row in enumerate(self.arr):
            for x, char in enumerate(row):
                surface.blit(self.charset.get_img(char), self.charset.char_size_to_pix((x, y)))

    def dump(self):
        """

        Returns:
            List[List[int]]: a copy of the frame buffer's grid
        """
        return list(map(copy, self.arr))


# TODO: given this class a better name.
class GameFrame(FrameBuffer):
    def __init__(self):
        self.objs = []

    def draw_char(self, c, x, y, height, width):
        self.objs.append(["char", [c, x, y, height, width]])

    def draw_tank(self, x, y, rotation, turret, fire, led, color):
        self.objs.append(["tank", [x, y, rotation, turret, fire, led, color]])

    def draw_crater(self, x, y, rotation, color):
        self.objs.append(["crater", [x, y, rotation, color]])

    def draw_sensors(self, x, y, rotation, turret, color, sensors):
        self.objs.append(["sensors", [x, y, rotation, turret, color, deepcopy(sensors)]])

    def get_obj_array(self):
        return self.objs



from __future__ import division

import pygame

PYGAME_SCREEN = None


def get_clock():
    return pygame.time.Clock()


class Display(object):
    def __init__(self, width, height, title=None):
        self.width = width
        self.height = height
        self.title = title

    def get_size(self):
        return self.width, self.height

    def get_keys(self):
        """Gets the keys that were press since last time you called this function.

        Returns:
            Key
        """
        raise NotImplementedError("This is an abstract method.")

    def update(self, frame_buffer):
        """Updates the display with the buffer given.

        Args:
            frame_buffer(FrameBuffer): The buffer to use when redrawing the screen.
        """
        raise NotImplementedError("This is an abstract method.")


class PyGameDisplay(Display):
    def __init__(self, width, height, title=None):
        super(PyGameDisplay, self).__init__(width, height, title)
        global PYGAME_SCREEN
        if not PYGAME_SCREEN:
            # Init pygame
            global pygame
            import pygame
            pygame.init()
            PYGAME_SCREEN = pygame.display.set_mode((self.width, self.height))
            if self.title:
                pygame.display.set_caption(self.title)
        else:
            raise Exception("You are trying the run two games at the time! I can't do that :(")

    def get_keys(self):
        keys = []
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                raise SystemExit()
            elif event.type == pygame.KEYDOWN and 0 < event.key < 256:
                char = chr(event.key)
                if pygame.key.get_mods() & pygame.KMOD_SHIFT:
                    # Shift was being pressed.
                    char = char.upper()
                keys += [char]
        return keys

    def update(self, frame_buffer):
        frame_buffer.draw_to_surface(PYGAME_SCREEN)
        pygame.display.flip()


class CharSet(object):
    def __init__(self, filename, char_width, char_height):
        self.char_width = char_width
        self.char_height = char_height
        self.img = pygame.image.load(filename)
        self.cache = {}

    def create_char_img(self, char):
        row = char % 16
        col = char // 16

        crop_dim = (row * self.char_width, col * self.char_height,
                    (row+1) *self.char_width, (col+1) * self.char_height)
        img = pygame.Surface((self.char_width, self.char_height), flags=pygame.SRCALPHA)
        img.blit(self.img, (0, 0), crop_dim)

        return img

    def get_img(self, char):
        if char not in self.cache:
            self.cache[char] = self.create_char_img(char)
        return self.cache[char]

    def pix_size_to_char(self, size):
        return size[0] // self.char_width, size[1] // self.char_height

    def char_size_to_pix(self, size):
        return size[0] * self.char_width, size[1] * self.char_height
